/*******************************************************************************
 * Copyright (c) 2004,2007 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.engine.toc;

import java.math.BigDecimal;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.birt.core.format.DateFormatter;
import org.eclipse.birt.core.format.NumberFormatter;
import org.eclipse.birt.core.format.StringFormatter;
import org.eclipse.birt.report.engine.api.ITOCTree;
import org.eclipse.birt.report.engine.api.TOCNode;
import org.eclipse.birt.report.engine.api.TOCStyle;
import org.eclipse.birt.report.engine.content.IStyle;
import org.eclipse.birt.report.engine.css.dom.CompositeStyle;
import org.eclipse.birt.report.engine.css.dom.StyleDeclaration;
import org.eclipse.birt.report.engine.css.engine.BIRTCSSEngine;
import org.eclipse.birt.report.engine.parser.StylePropertyMapping;
import org.eclipse.birt.report.model.api.ColorHandle;
import org.eclipse.birt.report.model.api.DimensionHandle;
import org.eclipse.birt.report.model.api.FontHandle;
import org.eclipse.birt.report.model.api.GroupHandle;
import org.eclipse.birt.report.model.api.ReportDesignHandle;
import org.eclipse.birt.report.model.api.ReportElementHandle;
import org.eclipse.birt.report.model.api.ReportItemHandle;
import org.eclipse.birt.report.model.api.StyleHandle;
import org.eclipse.birt.report.model.api.TOCHandle;
import org.eclipse.birt.report.model.elements.Style;
import org.w3c.dom.css.CSSValue;

import com.ibm.icu.util.TimeZone;
import com.ibm.icu.util.ULocale;

public class TOCTree implements ITOCTree
{

	private String format;

	private ULocale locale;
	
	private TimeZone timeZone = TimeZone.getDefault( );

	TOCTreeNode root;

	/** tocId, TOCNode map */
	private HashMap tocMapByID;

	private BIRTCSSEngine cssEngine;

	private ReportDesignHandle reportHandle;

	private Map<StyleKey, Styles> cachedStyles = new HashMap<StyleKey, Styles>( );
	
	public TOCTree( )
	{
	}

	public TOCTree( TOCTreeNode root, String format, ULocale locale,
			ReportDesignHandle reportHandle )
	{
		this( root, format, locale, null, reportHandle );
	}

	public TOCTree( TOCTreeNode root, String format, ULocale locale,
			TimeZone timeZone, ReportDesignHandle reportHandle )
	{
		this.root = root;
		this.format = format;
		if ( "viewer".equalsIgnoreCase( this.format ) )
		{
			this.format = "html";
		}
		this.locale = locale;
		this.reportHandle = reportHandle;
		this.cssEngine = new BIRTCSSEngine( );
		if ( timeZone != null )
		{
			this.timeZone = timeZone;
		}
	}

	public TOCNode findTOC( String tocNodeId )
	{
		initialize( );
		initializeLocale( );
		if ( root == null )
		{
			return null;
		}
		if ( tocMapByID == null )
		{
			generateIndex( );
		}
		TOCTreeNode treeNode = (TOCTreeNode) tocMapByID.get( tocNodeId );
		return refactNode( treeNode, format, locale );
	}

	public List findTOCByValue( Object tocValue )
	{
		if ( tocValue == null )
		{
			return null;
		}
		initialize( );
		initializeLocale( );
		if ( root == null )
		{
			return null;
		}
		List nodes = search( tocValue, locale );
		List result = new ArrayList( );
		for ( int i = 0; i < nodes.size( ); i++ )
		{
			TOCTreeNode treeNode = (TOCTreeNode) nodes.get( i );
			TOCNode node = refactNode( treeNode, format, locale );
			if ( node != null )
			{
				result.add( node );
			}
		}
		if ( result.size( ) == 0 )
		{
			return null;
		}
		return result;
	}

	public TOCTreeNode getTOCRoot( )
	{
		initialize( );
		return root;
	}

	public TOCNode getRoot( )
	{
		return extractTOCNode( root, null, format, locale );
	}

	public String getFormat( )
	{
		return format;
	}

	public void setFormat( String format )
	{
		this.format = format;
		if ( "viewer".equalsIgnoreCase( this.format ) )
		{
			this.format = "html";
		}
	}

	public ULocale getLocale( )
	{
		return locale;
	}

	public void setLocale( ULocale locale )
	{
		this.locale = locale;
	}
	
	public TimeZone getTimeZone( )
	{
		return this.timeZone;
	}

	public void setTimeZone( TimeZone timeZone )
	{
		if ( timeZone != null )
		{
			this.timeZone = timeZone;
		}
	}

	void setRoot( TOCTreeNode root )
	{
		this.root = root;
	}

	protected void initialize( )
	{
		if ( root == null )
		{
			root = new TOCTreeNode( );
		}
	}

	private List search( Object tocValue, ULocale locale )
	{
		Set nodes = new HashSet( );
		List keys = parse( tocValue, locale );
		search( nodes, keys, root );
		return Collections.list( Collections.enumeration( nodes ) );
	}

	private void search( Set result, List keys, TOCTreeNode rootNode )
	{
		for ( int i = 0; i < keys.size( ); i++ )
		{
			if ( equals( keys.get( i ), rootNode.getTOCValue( ) ) )
			{
				result.add( rootNode );
			}
		}
		List children = rootNode.getChildren( );
		for ( int i = 0; i < children.size( ); i++ )
		{
			search( result, keys, (TOCTreeNode) children.get( i ) );
		}
	}

	private boolean equals( Object searchKey, Object key )
	{
		assert searchKey != null;
		if ( key == null )
		{
			return false;
		}
		if ( searchKey instanceof String )
		{
			return searchKey.equals( key.toString( ) );
		}
		if ( searchKey instanceof Number && key instanceof Number )
		{
			BigDecimal searchDecimal = toBigDecimal( (Number) searchKey );
			BigDecimal decima2 = toBigDecimal( (Number) key );
			return searchDecimal.compareTo( decima2 ) == 0;
		}
		return searchKey.equals( key );
	}

	private BigDecimal toBigDecimal( Number value )
	{
		BigDecimal decimal;
		if ( value instanceof BigDecimal )
		{
			decimal = (BigDecimal) value;
		}
		else
		{
			decimal = new BigDecimal( value.toString( ) );
		}
		decimal = decimal.setScale( 20 );
		return decimal;
	}

	private TOCNode refactNode( TOCTreeNode node, String format, ULocale locale )
	{
		if ( node == null || node.isHidden( format ) )
		{
			return null;
		}
		if ( isFirstChildOfDummyGroup( node, format ) )
		{
			TOCTreeNode parent = (TOCTreeNode) node.getParent( );
			return extractTOCNode( parent, null, format, locale );
		}
		else
		{
			return extractTOCNode( node, null, format, locale );
		}
	}

	private boolean isFirstChildOfDummyGroup( TOCTreeNode node, String format )
	{
		TOCTreeNode parent = (TOCTreeNode) node.getParent( );
		if ( parent != null && isRootOfEmptyGroup( parent ) )
		{
			List children = parent.getChildren( );
			for ( int i = 0; i < children.size( ); i++ )
			{
				TOCTreeNode tmpNode = (TOCTreeNode) children.get( i );
				if ( tmpNode == node )
				{
					return true;
				}
				if ( !tmpNode.isHidden( format ) )
				{
					return false;
				}
			}
		}
		return false;
	}

	private void generateIndex( )
	{
		tocMapByID = new HashMap( );
		generateRootTOCIndex( );
	}

	/**
	 * extract the TOCTreeNode to the TOCNode.
	 * 
	 * @param root
	 * @param parent:
	 *            if the parent's value is null, that means the root is a top
	 *            level TOCTreeNode
	 * @param format
	 * @param locale
	 * @return
	 */
	private TOCNode extractTOCNode( TOCTreeNode root, TOCNode parent,
			String format, ULocale locale )
	{
		if ( root == null )
		{
			return null;
		}
		if ( root.isHidden( format ) )
		{
			return null;
		}
		TOCNode result = null;
		if ( !isRootOfEmptyGroup( root ) )
		{
			result = new TOCNode( root );
			if ( null != parent )
			{
				result.setParent( parent );
				parent.getChildren( ).add( result );
			}
			long elementId = root.getElementId( );
			if ( null != reportHandle )
			{
				Styles styles = getStyle( elementId, result );
				result.setTOCStyle( styles.tocStyle );
				result.setDisplayString( localizedValue( root.getTOCValue( ),
						styles.style, locale ) );
			}
			else
			{
				result.setDisplayString( localizedValue( root.getTOCValue( ),
						null, locale ) );
			}

		}
		List children = root.getChildren( );
		for ( int i = 0; i < children.size( ); i++ )
		{
			TOCTreeNode child = (TOCTreeNode) children.get( i );
			if ( null == result )
			{
				result = extractTOCNode( child, parent, format, locale );
			}
			else
			{
				extractTOCNode( child, result, format, locale );
			}
		}
		return result;
	}

	private Styles getStyle( long elementId, TOCNode result )
	{
		int level = getLevel( result );
		StyleKey styleKey = new StyleKey( elementId, level );
		Styles styles = cachedStyles.get( styleKey );
		if( styles == null )
		{
			IStyle style = getStyle( reportHandle, elementId, level );
			TOCStyle tocStyle = createTOCStyle( style );
			styles = new Styles(style, tocStyle);
			cachedStyles.put(styleKey, styles);
		}
		return styles;
	}

	private static class Styles{
		private IStyle style;
		private TOCStyle tocStyle;
		public Styles( IStyle style, TOCStyle tocStyle )
		{
			this.style = style;
			this.tocStyle = tocStyle;
		}
	}

	private static class StyleKey{
		private long elementId;
		private int level;
		public StyleKey(long elementId, int level )
		{
			this.elementId = elementId;
			this.level = level;
		}
		
		public boolean equals(Object obj)
		{
			if ( obj == this )
				return true;
			if ( !(obj instanceof StyleKey ))
				return false;
			StyleKey key = (StyleKey)obj;
			return elementId == key.elementId&& level == key.level;
		}
		
		public int hashCode( )
		{
			return (int)((elementId * 31) & 0xffff) + level;
		}
	}
	/**
	 * localize the value to a string.
	 * 
	 * @param value
	 * @param style
	 * @param locale
	 * @return
	 */
	private String localizedValue( Object value, IStyle style, ULocale locale )
	{
		if ( value == null )
		{
			return null;
		}
		if ( locale == null )
		{
			locale = ULocale.getDefault( );
		}
		String displayString = null;
		if ( value instanceof Number )
		{
			NumberFormatter fmt = null;
			if ( null == style )
			{
				fmt = new NumberFormatter( locale );
			}
			else
			{
				fmt = new NumberFormatter( style.getNumberFormat( ), locale );
			}
			displayString = fmt.format( (Number) value );
		}
		else if ( value instanceof java.util.Date )
		{
			DateFormatter fmt = null;
			if ( null == style )
			{
				fmt = new DateFormatter( locale, timeZone );
			}
			else
			{
				if ( value instanceof java.sql.Date )
				{
					fmt = new DateFormatter( style.getDateFormat( ), locale,
							timeZone );
				}
				else if ( value instanceof java.sql.Time )
				{
					fmt = new DateFormatter( style.getTimeFormat( ), locale,
							timeZone );
				}
				if( fmt == null )
				{
					fmt = new DateFormatter( style.getDateTimeFormat( ),
							locale, timeZone );
				}
			}
			displayString = fmt.format( (Date) value );
		}
		else if ( value instanceof String )
		{
			StringFormatter fmt = null;
			if ( null == style )
			{
				fmt = new StringFormatter( locale );
			}
			else
			{
				fmt = new StringFormatter( style.getStringFormat( ), locale );
			}
			displayString = fmt.format( (String) value );
		}
		else
		{
			displayString = value.toString( );
		}
		return displayString;
	}

	protected IStyle getStyle( ReportDesignHandle reportHandle, long elementId,
			int level )
	{
		ReportElementHandle handle = (ReportElementHandle) reportHandle
				.getElementByID( elementId );
		StyleHandle defaultStyleHandle = reportHandle
				.findStyle( TOCHandle.defaultTOCPrefixName + level );
		StyleHandle styleHandle = null;

		TOCHandle toc = null;
		if ( handle instanceof ReportItemHandle )
		{
			toc = ( (ReportItemHandle) handle ).getTOC( );
		}
		else if ( handle instanceof GroupHandle )
		{
			toc = ( (GroupHandle) handle ).getTOC( );
		}
		if ( toc != null )
		{
			String styleName = toc.getStyleName( );
			if ( styleName != null )
			{
				styleHandle = reportHandle.findStyle( styleName );
			}
		}

		IStyle defaultStyle = createStyle( defaultStyleHandle, cssEngine, true );
		IStyle style = createStyle( styleHandle, cssEngine, false );
		IStyle privateStyle = createPrivateStyle( toc, cssEngine );

		IStyle compositeStyle = new CompositeStyle( new CompositeStyle(
				defaultStyle, style ), privateStyle );
		return compositeStyle;
	}

	protected void addStylePropertyValue( IStyle style, String name,
			StyleHandle handle, boolean isDefault )
	{
		addStylePropertyValue( style, name, handle, false, isDefault );
	}

	protected void addStylePropertyValue( IStyle style, String name,
			StyleHandle handle, boolean isColorProperty, boolean isDefault )
	{
		Object value = null;
		int index = StylePropertyMapping.getPropertyID( name );

		if ( StylePropertyMapping.canInherit( name ) )
		{
			if ( handle != null )
			{
				if ( isColorProperty )
				{
					ColorHandle colorHandle = handle.getColorProperty( name );
					if ( colorHandle != null )
					{
						value = colorHandle.getStringValue( );
					}
				}
				else
				{
					value = handle.getProperty( name );
				}
			}
			if ( value == null && isDefault )
			{
				// bidi_hcg added arg
				value = StylePropertyMapping.getDefaultValue( name, reportHandle );
			}

			style.setCssText( index, value == null ? null : value.toString( ) );
		}
		else
		{
			if ( isDefault )
			{
				// bidi_hcg added arg
				value = StylePropertyMapping.getDefaultValue( name, reportHandle );
			}
			style.setCssText( index, value == null ? null : value.toString( ) );
		}

	}

	/**
	 * Creates the private styles
	 */
	protected IStyle createPrivateStyle( TOCHandle toc, BIRTCSSEngine cssEngine )
	{
		IStyle style = new StyleDeclaration( cssEngine );
		if ( null != toc )
		{
			// Background
			ColorHandle colorHandle = toc.getBackgroundColor( );
			style.setBackgroundColor( colorHandle == null ? null : colorHandle
					.getStringValue( ) );

			// Text related
			style.setTextAlign( toc.getTextAlign( ) );
			DimensionHandle dimensionHandle = toc.getTextIndent( );
			style.setTextIndent( dimensionHandle == null
					? null
					: dimensionHandle.getStringValue( ) );
			style.setTextTransform( toc.getTextTransform( ) );

			// Font related
			FontHandle fontHandle = toc.getFontFamily( );
			style.setFontFamily( fontHandle == null ? null : fontHandle
					.getStringValue( ) );
			colorHandle = toc.getColor( );
			style.setColor( colorHandle == null ? null : colorHandle
					.getStringValue( ) );
			dimensionHandle = toc.getFontSize( );
			style.setFontSize( dimensionHandle == null ? null : dimensionHandle
					.getStringValue( ) );
			style.setFontStyle( toc.getFontStyle( ) );
			style.setFontWeight( toc.getFontWeight( ) );
			style.setFontVariant( toc.getFontVariant( ) );

			// Text decoration
			style.setTextLineThrough( toc.getTextLineThrough( ) );
			style.setTextOverline( toc.getTextOverline( ) );
			style.setTextUnderline( toc.getTextUnderline( ) );

			// Border
			colorHandle = toc.getBorderBottomColor( );
			style.setBorderBottomColor( colorHandle == null
					? null
					: colorHandle.getStringValue( ) );
			style.setBorderBottomStyle( toc.getBorderBottomStyle( ) );
			dimensionHandle = toc.getBorderBottomWidth( );
			style.setBorderBottomWidth( dimensionHandle == null
					? null
					: dimensionHandle.getStringValue( ) );
			colorHandle = toc.getBorderLeftColor( );
			style.setBorderLeftColor( colorHandle == null ? null : colorHandle
					.getStringValue( ) );
			style.setBorderLeftStyle( toc.getBorderLeftStyle( ) );
			dimensionHandle = toc.getBorderLeftWidth( );
			style.setBorderLeftWidth( dimensionHandle == null
					? null
					: dimensionHandle.getStringValue( ) );
			colorHandle = toc.getBorderRightColor( );
			style.setBorderRightColor( colorHandle == null ? null : colorHandle
					.getStringValue( ) );
			style.setBorderRightStyle( toc.getBorderRightStyle( ) );
			dimensionHandle = toc.getBorderRightWidth( );
			style.setBorderRightWidth( dimensionHandle == null
					? null
					: dimensionHandle.getStringValue( ) );
			colorHandle = toc.getBorderTopColor( );
			style.setBorderTopColor( colorHandle == null ? null : colorHandle
					.getStringValue( ) );
			style.setBorderTopStyle( toc.getBorderTopStyle( ) );
			dimensionHandle = toc.getBorderTopWidth( );
			style.setBorderTopWidth( dimensionHandle == null
					? null
					: dimensionHandle.getStringValue( ) );

			// Format
			style.setCssText( StylePropertyMapping
					.getPropertyID( Style.STRING_FORMAT_PROP ), toc
					.getStringFormat( ) );
			style.setCssText( StylePropertyMapping
					.getPropertyID( Style.NUMBER_FORMAT_PROP ), toc
					.getNumberFormat( ) );
			style.setCssText( StylePropertyMapping
					.getPropertyID( Style.DATE_TIME_FORMAT_PROP ), toc
					.getDateTimeFormat( ) );
		}
		return style;
	}

	/**
	 * Creates Report default styles
	 */
	protected IStyle createStyle( StyleHandle handle, BIRTCSSEngine cssEngine,
			boolean isDefault )
	{

		IStyle style = new StyleDeclaration( cssEngine );

		if ( null != handle )
		{
			// Background
			addStylePropertyValue( style, Style.BACKGROUND_COLOR_PROP, handle,
					true, isDefault );
			addStylePropertyValue( style, Style.BACKGROUND_IMAGE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.BACKGROUND_POSITION_X_PROP,
					handle, isDefault );
			addStylePropertyValue( style, Style.BACKGROUND_POSITION_Y_PROP,
					handle, isDefault );
			addStylePropertyValue( style, Style.BACKGROUND_REPEAT_PROP, handle,
					isDefault );

			// Text related
			addStylePropertyValue( style, Style.TEXT_ALIGN_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.TEXT_INDENT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.LETTER_SPACING_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.LINE_HEIGHT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.ORPHANS_PROP, handle, isDefault );
			addStylePropertyValue( style, Style.TEXT_TRANSFORM_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.VERTICAL_ALIGN_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.WHITE_SPACE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.WIDOWS_PROP, handle, isDefault );
			addStylePropertyValue( style, Style.WORD_SPACING_PROP, handle,
					isDefault );

			// Section properties
			addStylePropertyValue( style, Style.DISPLAY_PROP, handle, isDefault );
			addStylePropertyValue( style, Style.MASTER_PAGE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.PAGE_BREAK_AFTER_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.PAGE_BREAK_BEFORE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.PAGE_BREAK_INSIDE_PROP, handle,
					isDefault );

			// Font related
			addStylePropertyValue( style, Style.FONT_FAMILY_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.COLOR_PROP, handle, true,
					isDefault );
			addStylePropertyValue( style, Style.FONT_SIZE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.FONT_STYLE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.FONT_WEIGHT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.FONT_VARIANT_PROP, handle,
					isDefault );

			// Text decoration
			addStylePropertyValue( style, Style.TEXT_LINE_THROUGH_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.TEXT_OVERLINE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.TEXT_UNDERLINE_PROP, handle,
					isDefault );

			// Border
			addStylePropertyValue( style, Style.BORDER_BOTTOM_COLOR_PROP,
					handle, true, isDefault );
			addStylePropertyValue( style, Style.BORDER_BOTTOM_STYLE_PROP,
					handle, isDefault );
			addStylePropertyValue( style, Style.BORDER_BOTTOM_WIDTH_PROP,
					handle, isDefault );
			addStylePropertyValue( style, Style.BORDER_LEFT_COLOR_PROP, handle,
					true, isDefault );
			addStylePropertyValue( style, Style.BORDER_LEFT_STYLE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.BORDER_LEFT_WIDTH_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.BORDER_RIGHT_COLOR_PROP,
					handle, true, isDefault );
			addStylePropertyValue( style, Style.BORDER_RIGHT_STYLE_PROP,
					handle, isDefault );
			addStylePropertyValue( style, Style.BORDER_RIGHT_WIDTH_PROP,
					handle, isDefault );
			addStylePropertyValue( style, Style.BORDER_TOP_COLOR_PROP, handle,
					true, isDefault );
			addStylePropertyValue( style, Style.BORDER_TOP_STYLE_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.BORDER_TOP_WIDTH_PROP, handle,
					isDefault );

			// Margin
			addStylePropertyValue( style, Style.MARGIN_TOP_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.MARGIN_LEFT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.MARGIN_BOTTOM_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.MARGIN_RIGHT_PROP, handle,
					isDefault );

			// Padding
			addStylePropertyValue( style, Style.PADDING_TOP_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.PADDING_LEFT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.PADDING_BOTTOM_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.PADDING_RIGHT_PROP, handle,
					isDefault );

			// Format
			addStylePropertyValue( style, Style.STRING_FORMAT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.NUMBER_FORMAT_PROP, handle,
					isDefault );
			addStylePropertyValue( style, Style.DATE_TIME_FORMAT_PROP, handle,
					isDefault );

			// bidi_hcg: Bidi related
			addStylePropertyValue( style, Style.TEXT_DIRECTION_PROP, handle, isDefault );
		}
		return style;
	}

	protected static int getLevel( TOCNode node )
	{
		TOCNode parent = node.getParent( );
		if ( parent != null )
		{
			return getLevel( parent ) + 1;
		}
		// Since the TOC start level as 0, we set -1 as root's level.
		return -1;
	}

	private void generateRootTOCIndex( )
	{
		tocMapByID.put( "/", root );
		generateTOCIndex( root );
	}

	/**
	 * add the TOC node into the map for search.
	 * 
	 * @param node
	 *            TOC cache.
	 * @param map
	 *            map contains the (id, TOC) pair.
	 */
	private void generateTOCIndex( TOCTreeNode node )
	{
		tocMapByID.put( node.getNodeID( ), node );
		Iterator iter = node.getChildren( ).iterator( );
		while ( iter.hasNext( ) )
		{
			TOCTreeNode child = (TOCTreeNode) iter.next( );
			generateTOCIndex( child );
		}
	}

	private boolean isRootOfEmptyGroup( TOCTreeNode node )
	{
		return node.isGroupRoot( ) && node.isBlank( );
	}

	private void initializeLocale( )
	{
		if ( locale == null )
		{
			locale = ULocale.getDefault( );
		}
	}

	private List parse( Object value, ULocale locale )
	{
		assert value != null;
		ArrayList result = new ArrayList( );
		result.add( value );
		if ( value instanceof String )
		{
			String stringValue = (String) value;
			Object parseValue = null;
			try
			{
				parseValue = new BigDecimal( stringValue );
			}
			catch ( NumberFormatException e )
			{
			}
			if ( parseValue != null )
			{
				result.add( parseValue );
				parseValue = null;
			}
			DateFormatter formatter = new DateFormatter( locale, timeZone );
			try
			{
				parseValue = formatter.parse( stringValue );
			}
			catch ( ParseException e )
			{
			}
			if ( parseValue != null )
			{
				result.add( parseValue );
			}
		}
		else
		{
			result.add( value.toString( ) );
		}
		return result;
	}

	private TOCStyle createTOCStyle( IStyle style )
	{
		TOCStyle tocStyle = new TOCStyle( );
		setTOCStyle( tocStyle, TOCStyle.BACKGROUND_ATTACHMENT, style,
				IStyle.STYLE_BACKGROUND_ATTACHMENT );
		setTOCStyle( tocStyle, TOCStyle.BACKGROUND_COLOR, style,
				IStyle.STYLE_BACKGROUND_COLOR );
		setTOCStyle( tocStyle, TOCStyle.BACKGROUND_IMAGE, style,
				IStyle.STYLE_BACKGROUND_IMAGE );
		setTOCStyle( tocStyle, TOCStyle.BACKGROUND_POSITION_X, style,
				IStyle.STYLE_BACKGROUND_POSITION_X );
		setTOCStyle( tocStyle, TOCStyle.BACKGROUND_POSITION_Y, style,
				IStyle.STYLE_BACKGROUND_POSITION_Y );
		setTOCStyle( tocStyle, TOCStyle.BACKGROUND_REPEAT, style,
				IStyle.STYLE_BACKGROUND_REPEAT );
		setTOCStyle( tocStyle, TOCStyle.BORDER_BOTTOM_COLOR, style,
				IStyle.STYLE_BORDER_BOTTOM_COLOR );
		setTOCStyle( tocStyle, TOCStyle.BORDER_TOP_COLOR, style,
				IStyle.STYLE_BORDER_TOP_COLOR );
		setTOCStyle( tocStyle, TOCStyle.BORDER_LEFT_COLOR, style,
				IStyle.STYLE_BORDER_LEFT_COLOR );
		setTOCStyle( tocStyle, TOCStyle.BORDER_RIGHT_COLOR, style,
				IStyle.STYLE_BORDER_RIGHT_COLOR );
		setTOCStyle( tocStyle, TOCStyle.BORDER_BOTTOM_WIDTH, style,
				IStyle.STYLE_BORDER_BOTTOM_WIDTH );
		setTOCStyle( tocStyle, TOCStyle.BORDER_TOP_WIDTH, style,
				IStyle.STYLE_BORDER_TOP_WIDTH );
		setTOCStyle( tocStyle, TOCStyle.BORDER_LEFT_WIDTH, style,
				IStyle.STYLE_BORDER_LEFT_WIDTH );
		setTOCStyle( tocStyle, TOCStyle.BORDER_RIGHT_WIDTH, style,
				IStyle.STYLE_BORDER_RIGHT_WIDTH );
		setTOCStyle( tocStyle, TOCStyle.BORDER_BOTTOM_STYLE, style,
				IStyle.STYLE_BORDER_BOTTOM_STYLE );
		setTOCStyle( tocStyle, TOCStyle.BORDER_TOP_STYLE, style,
				IStyle.STYLE_BORDER_TOP_STYLE );
		setTOCStyle( tocStyle, TOCStyle.BORDER_LEFT_STYLE, style,
				IStyle.STYLE_BORDER_LEFT_STYLE );
		setTOCStyle( tocStyle, TOCStyle.BORDER_RIGHT_STYLE, style,
				IStyle.STYLE_BORDER_RIGHT_STYLE );
		setTOCStyle( tocStyle, TOCStyle.CAN_SHRINK, style,
				IStyle.STYLE_CAN_SHRINK );
		setTOCStyle( tocStyle, TOCStyle.COLOR, style, IStyle.STYLE_COLOR );
		setTOCStyle( tocStyle, TOCStyle.DATE_FORMAT, style,
				IStyle.STYLE_DATE_FORMAT );
		setTOCStyle( tocStyle, TOCStyle.DISPLAY, style, IStyle.STYLE_DISPLAY );
		setTOCStyle( tocStyle, TOCStyle.FONT_FAMILY, style,
				IStyle.STYLE_FONT_FAMILY );
		setTOCStyle( tocStyle, TOCStyle.FONT_SIZE, style,
				IStyle.STYLE_FONT_SIZE );
		setTOCStyle( tocStyle, TOCStyle.FONT_STYLE, style,
				IStyle.STYLE_FONT_STYLE );
		setTOCStyle( tocStyle, TOCStyle.FONT_VARIANT, style,
				IStyle.STYLE_FONT_VARIANT );
		setTOCStyle( tocStyle, TOCStyle.FONT_WEIGHT, style,
				IStyle.STYLE_FONT_WEIGHT );
		setTOCStyle( tocStyle, TOCStyle.LETTER_SPACING, style,
				IStyle.STYLE_LETTER_SPACING );
		setTOCStyle( tocStyle, TOCStyle.LINE_HEIGHT, style,
				IStyle.STYLE_LINE_HEIGHT );
		setTOCStyle( tocStyle, TOCStyle.MARGIN_BOTTOM, style,
				IStyle.STYLE_MARGIN_BOTTOM );
		setTOCStyle( tocStyle, TOCStyle.MARGIN_LEFT, style,
				IStyle.STYLE_MARGIN_LEFT );
		setTOCStyle( tocStyle, TOCStyle.MARGIN_RIGHT, style,
				IStyle.STYLE_MARGIN_RIGHT );
		setTOCStyle( tocStyle, TOCStyle.MARGIN_TOP, style,
				IStyle.STYLE_MARGIN_TOP );
		setTOCStyle( tocStyle, TOCStyle.MASTER_PAGE, style,
				IStyle.STYLE_MASTER_PAGE );
		setTOCStyle( tocStyle, TOCStyle.NUMBER_FORMAT, style,
				IStyle.STYLE_NUMBER_FORMAT );
		setTOCStyle( tocStyle, TOCStyle.PADDING_BOTTOM, style,
				IStyle.STYLE_PADDING_BOTTOM );
		setTOCStyle( tocStyle, TOCStyle.PADDING_LEFT, style,
				IStyle.STYLE_PADDING_LEFT );
		setTOCStyle( tocStyle, TOCStyle.PADDING_RIGHT, style,
				IStyle.STYLE_PADDING_RIGHT );
		setTOCStyle( tocStyle, TOCStyle.PADDING_TOP, style,
				IStyle.STYLE_PADDING_TOP );
		setTOCStyle( tocStyle, TOCStyle.PAGE_BREAK_AFTER, style,
				IStyle.STYLE_PAGE_BREAK_AFTER );
		setTOCStyle( tocStyle, TOCStyle.PAGE_BREAK_BEFORE, style,
				IStyle.STYLE_PAGE_BREAK_BEFORE );
		setTOCStyle( tocStyle, TOCStyle.PAGE_BREAK_INSIDE, style,
				IStyle.STYLE_PAGE_BREAK_INSIDE );
		setTOCStyle( tocStyle, TOCStyle.SHOW_IF_BLANK, style,
				IStyle.STYLE_SHOW_IF_BLANK );
		setTOCStyle( tocStyle, TOCStyle.STRING_FORMAT, style,
				IStyle.STYLE_STRING_FORMAT );
		setTOCStyle( tocStyle, TOCStyle.TEXT_ALIGN, style,
				IStyle.STYLE_TEXT_ALIGN );
		setTOCStyle( tocStyle, TOCStyle.TEXT_INDENT, style,
				IStyle.STYLE_TEXT_INDENT );
		setTOCStyle( tocStyle, TOCStyle.TEXT_LINE_THROUGH, style,
				IStyle.STYLE_TEXT_LINETHROUGH );
		setTOCStyle( tocStyle, TOCStyle.TEXT_OVERLINE, style,
				IStyle.STYLE_TEXT_OVERLINE );
		setTOCStyle( tocStyle, TOCStyle.TEXT_TRANSFORM, style,
				IStyle.STYLE_TEXT_TRANSFORM );
		setTOCStyle( tocStyle, TOCStyle.TEXT_UNDERLINE, style,
				IStyle.STYLE_TEXT_UNDERLINE );
		setTOCStyle( tocStyle, TOCStyle.VERTICAL_ALIGN, style,
				IStyle.STYLE_VERTICAL_ALIGN );
		setTOCStyle( tocStyle, TOCStyle.VISIBLE_FORMAT, style,
				IStyle.STYLE_VISIBLE_FORMAT );
		setTOCStyle( tocStyle, TOCStyle.WHITE_SPACE, style,
				IStyle.STYLE_WHITE_SPACE );
		setTOCStyle( tocStyle, TOCStyle.WORD_SPACING, style,
				IStyle.STYLE_WORD_SPACING );
		return tocStyle;
	}

	private void setTOCStyle( TOCStyle tocStyle, String name, IStyle style,
			int index )
	{
		CSSValue value = style.getProperty( index );
		if ( value != null )
		{
			tocStyle.setProperty( name, value.getCssText( ) );
		}
	}
}
